@@ -47,7 +47,7 @@ void Rprintf(const char *, ...);
 #define BUFSIZE 0x10000
 
 /* seems not ever to be used? */
-char errmsg[CL_MAX_LINE_LENGTH];
+/* char errmsg[CL_MAX_LINE_LENGTH]; */
 
 
 
